/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "livedumper.hpp"
#include "log.hpp"

#include <unistd.h>

#include <boost/exception/diagnostic_information.hpp>
#include <boost/exception/info.hpp>
#include <boost/exception_ptr.hpp>
#include <exception>

#if defined(__arm__) || defined(__i386__)
#define ELF_TYPE Elf32
#elif defined(__x86_64__) || defined(__aarch64__)
#define ELF_TYPE Elf64
#endif

void help(const char *app_name) {
	std::cout << "Usage: " << std::endl
	          << "  " << app_name << " [-f file_name] [-P fd] <pid>" << std::endl
	          << std::endl
	          << "  -f <file_name>   output path" << std::endl
	          << "  -P <fd>          descriptor nr" << std::endl
	          << "  -m               save minicore" << std::endl;
}

int main(int argc, char *argv[]) {
	int opt;
	int prstatus_fd = -1;

	std::string output_file;
	bool minicore = false;

	while ((opt = getopt(argc, argv, "hP:f:m")) != -1) {
		switch (opt) {
		case 'P':
			prstatus_fd = atoi(optarg);
			break;
		case 'h':
			help(argv[0]);
			exit(EXIT_SUCCESS);
		case 'f':
			output_file = optarg;
			break;
		case 'm':
			minicore = true;
			break;
		default:
			help(argv[0]);
			exit(EXIT_FAILURE);
		}
	}

	if (argc <= optind) {
		std::cout << "PID not provided" << std::endl;
		return -1;
	}

	std::string pid_str = std::string(argv[optind]);

	if (output_file.empty())
		output_file = "livecore." + pid_str;

	LiveDumper<ELF_TYPE> dumper(std::stoi(pid_str));

	if (prstatus_fd > 0)
		dumper.setPrstatusFd(prstatus_fd);

	try {
		return dumper.DumpCore(output_file, minicore) ? EXIT_SUCCESS : EXIT_FAILURE;
	} catch (std::exception &e) {
		logger.log_error("%s", e.what());
		std::cout << e.what() << std::endl;
		return EXIT_FAILURE;
	} catch (...) {
		logger.log_error("unspecified error");
		std::cout << "unspecified error" << std::endl;
		return EXIT_FAILURE;
	}
}
